@@ -1,6 +1,6 @@
 package Net::OSCAR::Callbacks;
 
-$VERSION = 0.56;
+$VERSION = 0.57;
 
 use strict;
 use vars qw($VERSION);
@@ -120,8 +120,7 @@ sub process_snac($$) {
 		} elsif($conntype == CONNTYPE_CHAT) {
 			$connection->ready();
 
-			# We don't get the 0x0E/0x02 callback for exchange 5 chats until it's too late
-			$session->callback_chat_joined($connection->name, $connection) if $connection->exchange == 5;
+			$session->callback_chat_joined($connection->name, $connection) unless $connection->{sent_joined}++;
 		}
 	} elsif($subtype == 0x1) {
 		$subtype = $reqid >> 16;
@@ -219,9 +218,14 @@ sub process_snac($$) {
 		$connection->log_print(OSCAR_DBG_DEBUG, "Got incoming IM.");
 		my($from, $msg, $away, $chat, $chaturl) = $session->im_parse($data);
 		if($from) {
-			if($chat) {
+			# Ignore invites for chats that we're already in
+			if($chat and not
+				grep { $_->{url} eq $chaturl }
+					 grep { $_->{conntype} == CONNTYPE_CHAT }
+						@{$session->{connections}}
+			) {
 				$session->callback_chat_invite($from, $msg, $chat, $chaturl);
-			} else {
+			} elsif(!$chat) {
 				$session->callback_im_in($from, $msg, $away);
 			}
 		}
@@ -388,7 +392,7 @@ sub process_snac($$) {
 		my($tlvcount) = unpack("n", substr($data, 0, 2, ""));
 		my $tlv = tlv_decode($data);
 
-		$session->callback_chat_joined($connection->{name}, $connection) unless $connection->exchange == 5;
+		$session->callback_chat_joined($connection->{name}, $connection) unless $connection->{sent_joined}++;
 
 		my $occupants = 0;
 		($occupants) = unpack("n", $tlv->{0x6F}) if $tlv->{0x6F};
